
import jwt
import logging
import re
import time

from qtoggleserver import system
from qtoggleserver.conf import settings
from qtoggleserver.core.device import attrs as core_device_attrs


JWT_ISS = 'qToggle'
JWT_ALG = 'HS256'

ORIGIN_DEVICE = 'device'
ORIGIN_CONSUMER = 'consumer'

_AUTH_TOKEN_RE = re.compile(r'^Bearer\s+([a-z0-9_.-]+)$', re.IGNORECASE)

logger = logging.getLogger(__name__)


class AuthError(Exception):
    pass


def make_auth_header(origin, username, password_hash):
    claims = {
        'iss': JWT_ISS,
        'ori': origin
    }

    if username:
        claims['usr'] = username

    if system.date.has_date_support() and system.date.has_real_date_time():
        claims['iat'] = int(time.time())

    token = jwt.encode(claims, key=password_hash or '', algorithm=JWT_ALG)

    return 'Bearer {}'.format(token.decode())


def parse_auth_header(auth, origin, password_hash_func, require_usr=True):
    m = _AUTH_TOKEN_RE.match(auth)
    if not m:
        raise AuthError('invalid authorization header')

    # decode but don't validate token yet
    token = m.group(1)

    try:
        payload = jwt.decode(token, algorithms=[JWT_ALG], verify=False)

    except jwt.exceptions.InvalidTokenError as e:
        raise AuthError('invalid JWT: {}'.format(e)) from e

    # validate claims
    if payload.get('iss') != JWT_ISS:
        raise AuthError('missing or invalid iss claim in JWT')

    if payload.get('ori') != origin:
        raise AuthError('missing or invalid ori claim in JWT')

    iat = payload.get('iat')
    if (iat is not None) and system.date.has_date_support() and system.date.has_real_date_time():
        delta = time.time() - iat
        if abs(delta) > settings.core.max_client_time_skew:
            raise AuthError('JWT too old or too much in the future')

    usr = payload.get('usr')
    if require_usr:
        if not usr or not isinstance(usr, str):
            raise AuthError('missing or invalid usr claim in JWT')

    # validate username & signature
    password_hash = password_hash_func(usr)
    if not password_hash:
        raise AuthError('unknown usr in JWT: {}'.format(usr))

    # decode again to verify signature
    try:
        jwt.decode(token, key=password_hash, algorithms=[JWT_ALG], verify=True)

    except jwt.exceptions.InvalidSignatureError as e:
        raise AuthError('invalid JWT signature') from e

    except jwt.exceptions.InvalidTokenError as e:
        raise AuthError('invalid JWT: {}'.format(e)) from e

    return usr


def consumer_password_hash_func(usr):
    if usr == 'admin':
        return core_device_attrs.admin_password_hash

    elif usr == 'normal':
        return core_device_attrs.normal_password_hash

    elif usr == 'viewonly':
        return core_device_attrs.viewonly_password_hash

    else:
        return None




from qtoggleserver import system
from qtoggleserver.core import api as core_api
from qtoggleserver.core import device as core_device
from qtoggleserver.core import main
from qtoggleserver.core.api import schema as core_api_schema
from qtoggleserver.core.device import attrs as core_device_attrs
from qtoggleserver.core.device import events as core_device_events


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def get_device(request):
    return core_device_attrs.to_json()


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def patch_device(request, params):
    def unexpected_field_msg(field):
        if field in core_device_attrs.ATTRDEFS:
            return 'attribute not modifiable: {field}'

        else:
            return 'no such attribute: {field}'

    core_api_schema.validate(params, core_device_attrs.get_schema(), unexpected_field_msg=unexpected_field_msg)

    try:
        result = core_device_attrs.set_attrs(params)

    except Exception as e:
        raise core_api.APIError(500, str(e)) from e

    # noinspection PyTypeChecker
    if isinstance(result, str):  # API error
        raise core_api.APIError(400, result)

    core_device.save()
    core_device_events.trigger_update()

    if result:  # reboot required
        main.loop.call_later(2, system.reboot)




import asyncio

from qtoggleserver.conf import settings
from qtoggleserver.core import api as core_api
from qtoggleserver.core import main
from qtoggleserver.core import ports as core_ports
from qtoggleserver.core import vports as core_vports
from qtoggleserver.core.api import schema as core_api_schema
from qtoggleserver.utils import json as json_utils


@core_api.api_call(core_api.ACCESS_LEVEL_VIEWONLY)
async def get_ports(request):
    return [await port.to_json() for port in sorted(core_ports.all_ports(), key=lambda p: p.get_id())]


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def patch_port(request, port_id, params):
    port = core_ports.get(port_id)
    if port is None:
        raise core_api.APIError(404, 'no such port')

    def unexpected_field_msg(field):
        if field in port.get_non_modifiable_attrs():
            return 'attribute not modifiable: {field}'

        else:
            return 'no such attribute: {field}'

    core_api_schema.validate(params, await port.get_schema(), unexpected_field_msg=unexpected_field_msg)

    # step validation
    for name, value in params.items():
        attrdef = port.ATTRDEFS[name]
        step = attrdef.get('step')
        _min = attrdef.get('min')
        if None not in (step, _min) and step != 0 and (value - _min) % step:
            raise core_api.APIError(400, 'invalid field: {}'.format(name))

    errors_by_name = {}

    async def set_attr(attr_name, attr_value):
        core_api.logger.debug('setting attribute %s = %s on %s', attr_name, json_utils.dumps(attr_value), port)

        try:
            await port.set_attr(attr_name, attr_value)

        except Exception as e:
            errors_by_name[attr_name] = e

    if params:
        await asyncio.wait([set_attr(name, value) for name, value in params.items()])

    if errors_by_name:
        name, error = next(iter(errors_by_name.items()))

        if isinstance(error, core_api.APIError):
            raise error

        elif isinstance(error, core_ports.InvalidAttributeValue):
            raise core_api.APIError(400, 'invalid field: {}'.format(name))

        elif isinstance(error, core_ports.PortTimeout):
            raise core_api.APIError(504, 'port timeout')

        elif isinstance(error, core_ports.PortError):
            raise core_api.APIError(502, 'port error: {}'.format(error))

        else:
            # transform any unhandled exception into APIError(500)
            raise core_api.APIError(500, str(error))

    await port.save()


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def post_ports(request, params):
    core_api_schema.validate(params, core_api_schema.POST_PORTS)

    port_id = params['id']
    port_type = params['type']
    mi = params.get('min')
    ma = params.get('max')
    integer = params.get('integer')
    step = params.get('step')
    choices = params.get('choices')

    if core_ports.get(port_id):
        raise core_api.APIError(400, 'duplicate port')

    if len(core_vports.all_settings()) >= settings.core.virtual_ports:
        raise core_api.APIError(400, 'too many ports')

    core_vports.add(port_id, port_type, mi, ma, integer, step, choices)
    port = await core_ports.load_one('qtoggleserver.core.vports.VirtualPort',
                                     {'port_id': port_id, 'typ': port_type, 'mi': mi, 'ma': ma,
                                      'integer': integer, 'step': step, 'choices': choices})

    # a virtual port is enabled by default
    await port.enable()
    await port.save()

    return await port.to_json()


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def delete_port(request, port_id):
    port = core_ports.get(port_id)
    if not port:
        raise core_api.APIError(404, 'no such port')

    if not isinstance(port, core_vports.VirtualPort):
        raise core_api.APIError(400, 'port not removable')

    if not port.remove():
        raise core_api.APIError(500, 'failed to remove port')

    if not core_vports.remove(port_id):
        raise core_api.APIError(500, 'failed to remove virtual port settings')


@core_api.api_call(core_api.ACCESS_LEVEL_VIEWONLY)
async def get_port_value(request, port_id):
    port = core_ports.get(port_id)
    if port is None:
        raise core_api.APIError(404, 'no such port')

    if not port.is_enabled():
        return

    # TODO
    # given the fact that get_value() simply returns last cached read value,
    # and the fact that API specs indicate that 502/504 be returned by GET /port/[id]/value in case of errors
    # we should remember the last error generated by read_value() and return it here, if any

    return port.get_value()


@core_api.api_call(core_api.ACCESS_LEVEL_NORMAL)
async def patch_port_value(request, port_id, params):
    port = core_ports.get(port_id)
    if port is None:
        raise core_api.APIError(404, 'no such port')

    core_api_schema.validate(params, await port.get_value_schema(), invalid_request_msg='invalid value')

    value = params

    # step validation
    step = await port.get_attr('step')
    _min = await port.get_attr('min')
    if None not in (step, _min) and step != 0 and (value - _min) % step:
        raise core_api.APIError(400, 'invalid field: value')

    if not port.is_enabled():
        raise core_api.APIError(400, 'port disabled')

    if not await port.is_writable():
        raise core_api.APIError(400, 'read-only port')

    try:
        await port.set_value(value, reason=core_ports.CHANGE_REASON_API)

    except core_ports.PortTimeout as e:
        raise core_api.APIError(504, 'port timeout') from e

    except core_ports.PortError as e:
        raise core_api.APIError(502, 'port error: {}'.format(e)) from e

    except core_api.APIError:
        raise

    except Exception as e:
        # transform any unhandled exception into APIError(500)
        raise core_api.APIError(500, str(e)) from e

    await main.update()


@core_api.api_call(core_api.ACCESS_LEVEL_NORMAL)
async def post_port_sequence(request, port_id, params):
    port = core_ports.get(port_id)
    if port is None:
        raise core_api.APIError(404, 'no such port')

    core_api_schema.validate(params, core_api_schema.POST_PORT_SEQUENCE)

    values = params['values']
    delays = params['delays']
    repeat = params['repeat']

    if len(values) != len(delays):
        raise core_api.APIError(400, 'invalid field: delays')

    value_schema = await port.get_value_schema()
    step = await port.get_attr('step')
    _min = await port.get_attr('min')
    for value in values:
        core_api_schema.validate(value, value_schema, invalid_request_msg='invalid field: values')

        # step validation
        if None not in (step, _min) and step != 0 and (value - _min) % step:
            raise core_api.APIError(400, 'invalid field: values')

    if not port.is_enabled():
        raise core_api.APIError(400, 'port disabled')

    if not await port.is_writable():
        raise core_api.APIError(400, 'read-only port')

    if await port.get_attr('expression'):
        raise core_api.APIError(400, 'port with expression')

    try:
        await port.set_sequence(values, delays, repeat)

    except Exception as e:
        # transform any unhandled exception into APIError(500)
        raise core_api.APIError(500, str(e)) from e




import logging

from qtoggleserver.core.events import BaseEventHandler


logger = logging.getLogger(__name__)


class DummyEventHandler(BaseEventHandler):
    async def on_event(self, event):
        logger.debug('handling %s', event)

    async def on_value_change(self, event, port, old_value, new_value, attrs):
        logger.debug('handling value-change for %s', port)

    async def on_port_update(self, event, port, old_attrs, new_attrs, changed_attrs, added_attrs, removed_attrs):
        logger.debug('handling port-update for %s', port)

    async def on_port_add(self, event, port, attrs):
        logger.debug('handling port-add for %s', port)

    async def on_port_remove(self, event, port, attrs):
        logger.debug('handling port-remove for %s', port)

    async def on_device_update(self, event, old_attrs, new_attrs, changed_attrs, added_attrs, removed_attrs):
        logger.debug('handling device-update')

    async def on_slave_device_update(self, event, slave, old_attrs, new_attrs, changed_attrs, added_attrs, removed_attrs):
        logger.debug('handling slave-device-update for %s', slave)

    async def on_slave_device_add(self, event, slave, attrs):
        logger.debug('handling slave-device-add for %s', slave)

    async def on_slave_device_remove(self, event, slave, attrs):
        logger.debug('handling slave-device-remove for %s', slave)




import logging
import subprocess


logger = logging.getLogger(__name__)


def run_get_cmd(get_cmd, cmd_name=None, log_values=True, exc_class=None, required_fields=None):
    exc_class = exc_class or Exception

    try:
        config = subprocess.check_output(get_cmd, stderr=subprocess.STDOUT, shell=True)

    except Exception as e:
        raise exc_class('{} get command failed: {}'.format(cmd_name or get_cmd, e)) from e

    config = config.strip().decode()
    config_lines = config.split('\n')
    config_dict = {}
    for line in config_lines:
        line = line.strip()
        if not line:
            continue

        parts = line.split('=', 1)
        if len(parts) == 1:
            parts.append('')

        key, value = parts
        key = key.lower()[3:]  # strip leading "QS_"
        if value.startswith('"'):
            value = value[1:]
        if value.endswith('"') and not value.endswith('\\"'):
            value = value[:-1]

        config_dict[key] = value

    if cmd_name:
        if log_values:
            values_str = ', '.join(['{} = "{}"'.format(k, v) for k, v in sorted(config_dict.items())])
            logger.debug('got %s: %s', cmd_name, values_str)

        else:
            logger.debug('got %s', cmd_name)

    for field in required_fields or []:
        if field not in config_dict:
            msg = 'missing {} field'.format(field)
            if cmd_name:
                msg = 'invalid {}: {}'.format(cmd_name, msg)

            raise exc_class(msg)

    return config_dict


def run_set_cmd(set_cmd, cmd_name=None, log_values=True, exc_class=None, **config):
    env = {'QS_{}'.format(k.upper()): v for k, v in config.items()}

    exc_class = exc_class or Exception

    try:
        subprocess.check_output(set_cmd, env=env, stderr=subprocess.STDOUT, shell=True)

    except Exception as e:
        raise exc_class('{} set command failed: {}'.format(cmd_name or set_cmd, e)) from e

    if cmd_name:
        if log_values:
            values_str = ', '.join(['{} = "{}"'.format(k, v) for k, v in sorted(config.items())])
            logger.debug('%s set to: %s', cmd_name, values_str)

        else:
            logger.debug('%s set', cmd_name)




import abc
import logging

from qtoggleserver import utils
from qtoggleserver.conf import settings
from qtoggleserver.conf import utils as conf_utils
from qtoggleserver.utils import json as json_utils


logger = logging.getLogger(__name__)

_driver = None


class BaseDriver(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def query(self, collection, fields, filt, limit):
        return []

    @abc.abstractmethod
    def insert(self, collection, record):
        return '1'  # returns the inserted record id

    @abc.abstractmethod
    def update(self, collection, record_part, filt):
        return 0  # returns the number of updated records

    @abc.abstractmethod
    def replace(self, collection, _id, record, upsert):
        return False  # returns True if replaced

    @abc.abstractmethod
    def remove(self, collection, filt):
        return 0  # returns the number of removed records

    @abc.abstractmethod
    def close(self):
        pass


def _get_driver():
    global _driver

    if _driver is None:
        driver_args = conf_utils.obj_to_dict(settings.persist)
        driver_class_path = driver_args.pop('driver')

        try:
            logger.debug('loading persistence driver %s', driver_class_path)
            driver_class = utils.load_attr(driver_class_path)
            _driver = driver_class(**driver_args)

        except Exception as e:
            logger.error('failed to load persistence driver %s: %s', driver_class_path, e, exc_info=True)

            raise

    return _driver


def query(collection, fields=None, filt=None, limit=None):
    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug('querying %s (%s) where %s', collection, json_utils.dumps(fields) if fields else 'all fields',
                     json_utils.dumps(filt))

    return _get_driver().query(collection, fields, filt or {}, limit)


def get(collection, _id):
    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug('getting record with id %s from %s', _id, collection)

    records = list(_get_driver().query(collection, fields=None, filt={'id': _id}, limit=1))
    if records:
        return records[0]

    return None


def get_value(name, default=None):
    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug('getting value of %s', name)

    records = list(_get_driver().query(name, fields=None, filt={}, limit=2))
    if len(records) > 1:
        logger.warning('more than one record found in single-value collection %s', name)

        record = records[0]

    elif len(records) > 0:
        record = records[0]

    else:
        return default

    return record['value']


def set_value(name, value):
    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug('setting %s to %s', name, json_utils.dumps(value))

    record = {'value': value}
    _get_driver().replace(name, '', record, upsert=True)


def insert(collection, record):
    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug('inserting %s into %s', json_utils.dumps(record), collection)

    return _get_driver().insert(collection, record)


def update(collection, record_part, filt=None):
    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug('updating %s where %s with %s', collection, json_utils.dumps(filt or {}),
                     json_utils.dumps(record_part))

    count = _get_driver().update(collection, record_part, filt or {})

    logger.debug('modified %s records in %s', count, collection)

    return count


def replace(collection, _id, record, upsert=True):
    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug('replacing record with id %s with %s in %s', _id, json_utils.dumps(record), collection)

    record = dict(record, id=_id)  # make sure the new record contains the id field
    replaced = _get_driver().replace(collection, _id, record, upsert)

    if replaced:
        logger.debug('replaced record with id %s in %s', _id, collection)

    return replaced


def remove(collection, filt=None):
    if logger.getEffectiveLevel() <= logging.DEBUG:
        logger.debug('removing from %s where %s', collection, json_utils.dumps(filt or {}))

    count = _get_driver().remove(collection, filt or {})

    logger.debug('removed %s records from %s', count, collection)

    return count


def close():
    logger.debug('closing')

    return _get_driver().close()




import asyncio

from tornado import process as tornado_process


async def call_subprocess(args, stdin_data=None):
    p = tornado_process.Subprocess(args,
                                   stdin=tornado_process.Subprocess.STREAM,
                                   stdout=tornado_process.Subprocess.STREAM,
                                   stderr=tornado_process.Subprocess.STREAM)

    if stdin_data:
        await p.stdin.write(stdin_data)
        p.stdin.close()

    exit_future = p.wait_for_exit(raise_error=False)
    stdout_future = p.stdout.read_until_close()
    stderr_future = p.stderr.read_until_close()

    await asyncio.wait({exit_future, stdout_future, stderr_future})

    return exit_future.result(), stdout_future.result(), stderr_future.result()




import logging
import asyncio


from qtoggleserver import utils
from qtoggleserver.conf import settings

from .base import BaseEventHandler


logger = logging.getLogger(__name__)

_registered_handlers = []


def register_handler(handler):
    _registered_handlers.append(handler)


def init():
    for handler_args in settings.event_handlers:
        handler_class_path = handler_args.pop('driver')

        try:
            logger.debug('loading event handler %s', handler_class_path)
            handler_class = utils.load_attr(handler_class_path)
            handler = handler_class(**handler_args)

        except Exception as e:
            logger.error('failed to load event handler %s: %s', handler_class_path, e, exc_info=True)

        else:
            _registered_handlers.append(handler)


def handle_event(event):
    for handler in _registered_handlers:
        asyncio.create_task(handler.handle_event(event))




import os

from qtoggleserver.core import ports
from qtoggleserver.utils import json as json_utils


class GPIO(ports.Port):
    TYPE = ports.TYPE_BOOLEAN

    ADDITIONAL_ATTRDEFS = {
        'output': {
            'display_name': 'Is Output',
            'description': 'Controls the port direction.',
            'type': 'boolean',
            'modifiable': True
        }
    }

    BASE_PATH = '/sys/class/gpio'

    def __init__(self, no, def_value=None, def_output=None):
        self._no = no
        self._def_value = def_value
        self._def_output = def_output

        self._val_file = None
        self._dir_file = None

        super().__init__(port_id='gpio{}'.format(no))

    async def handle_enable(self):
        try:
            (self._val_file, self._dir_file) = self._configure()

        except Exception as e:
            self.error('failed to configure %s: %s', self, e)

            raise

        if self._def_output is not None:
            await self.attr_set_output(self._def_output)

    async def read_value(self):
        self._val_file.seek(0)

        return self._val_file.read(1) == '1'

    async def write_value(self, value):
        self._val_file.seek(0)

        if value:
            value = '1'

        else:
            value = '0'

        self.debug('writing %s to "%s"', json_utils.dumps(value), self._val_file.name)
        self._val_file.write(value)
        self._val_file.flush()

    async def attr_is_writable(self):
        return self._is_output()

    async def attr_set_output(self, output):
        if not self._dir_file:
            return

        self._dir_file.seek(0)

        if output:
            text = 'out'

        else:
            text = 'in'

        self.debug('writing "%s" to "%s"', text, self._dir_file.name)
        self._dir_file.write(text)
        self._dir_file.flush()

        if output and self._def_value is not None:
            await self.write_value(self._def_value)

    async def attr_is_output(self):
        return self._is_output()

    def _is_output(self):
        if not self._dir_file:
            return False

        self._dir_file.seek(0)

        return self._dir_file.read(3) == 'out'

    def _configure(self):
        path = os.path.join(self.BASE_PATH, self.get_id())

        if not os.path.exists(path):
            self.debug('exporting %s', self)
            with open(os.path.join(self.BASE_PATH, 'export'), 'w') as f:
                f.write(str(self._no))

        return (open(os.path.join(path, 'value'), 'r+'),
                open(os.path.join(path, 'direction'), 'r+'))




import os.path
import shutil

from setuptools import setup, find_namespace_packages, Command
from setuptools.command.sdist import sdist

try:
    from wheel.bdist_wheel import bdist_wheel

except ImportError:
    bdist_wheel = None

import qtoggleserver.version

try:
    import fastentrypoints

except ImportError:
    pass

try:
    import setupnovernormalize

except ImportError:
    pass


name = 'qtoggleserver'
version = qtoggleserver.version.VERSION


class UIMakeMixin:
    def __init__(self):
        self.root_path = os.getcwd()
        self.build_path = None

    def copy_frontend_files(self):
        dest_dir = os.path.join(self.build_path, name, 'frontend')

        src_dir = os.path.join(self.root_path, 'frontend', 'dist')
        if not os.path.exists(src_dir):
            # Running from a sdist
            src_dir = os.path.join(self.root_path, 'qtoggleserver', 'frontend')

        # Remove frontend dir created by in-source symlink
        shutil.rmtree(dest_dir)
        shutil.copytree(src_dir, dest_dir)


class SdistCommand(sdist, UIMakeMixin):
    def __init__(self, *args, **kwargs):
        sdist.__init__(self, *args, **kwargs)
        UIMakeMixin.__init__(self)

    def run(self):
        self.build_path = os.path.join(self.root_path, self.distribution.get_fullname())

        sdist.run(self)

    def make_release_tree(self, base_dir, files):
        sdist.make_release_tree(self, base_dir, files)

        self.copy_frontend_files()


if bdist_wheel:
    class BdistWheelCommand(bdist_wheel, UIMakeMixin):
        def __init__(self, *args, **kwargs):
            bdist_wheel.__init__(self, *args, **kwargs)
            UIMakeMixin.__init__(self)

        def egg2dist(self, egginfo_path, distinfo_path):
            self.build_path = self.bdist_dir
            self.copy_frontend_files()

            bdist_wheel.egg2dist(self, egginfo_path, distinfo_path)

else:
    class BdistWheelCommand(Command):
        user_options = []

        def initialize_options(self):
            pass

        def finalize_options(self):
            pass

        def run(self):
            raise NotImplementedError()


class CleanCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        os.system('rm -vrf ./build ./dist ./*.egg-info')


setup(
    name=name,
    version=version,

    description='A fully fledged qToggle Python implementation',

    author='Calin Crisan',
    author_email='ccrisan@gmail.com',

    license='Apache 2.0',

    packages=find_namespace_packages(include=name + '/*'),
    namespace_packages=['qtoggleserver'],

    install_requires=[
        'jinja2',
        'jsonpointer',
        'jsonschema',
        'pyhocon',
        'pyjwt',
        'pytz',
        'tornado',
    ],

    package_data={
        name: [
            'frontend/*.css',
            'frontend/*.js',
            'frontend/*.map',
            'frontend/font/*.*',
            'frontend/img/*.*',
            'frontend/html/*.*',
        ]
    },

    entry_points={
        'console_scripts': [
            'qtoggleserver=qtoggleserver.commands.server:execute',
            'qtoggleshell=qtoggleserver.commands.shell:execute'
        ],
    },

    cmdclass={
        'sdist': SdistCommand,
        'bdist_wheel': BdistWheelCommand,
        'clean': CleanCommand
    }
)




import asyncio
import logging
import time

from qtoggleserver.conf import settings
from qtoggleserver.utils import json as json_utils


logger = logging.getLogger(__name__)
memory_logs = None

loop = asyncio.get_event_loop()

_update_loop_task = None
_running = True
_ready = False
_last_time = 0
_force_eval_expressions = False


async def update():
    from . import ports
    from . import sessions

    global _last_time

    changed_set = set()
    change_reasons = {}

    now = int(time.time())
    time_changed = False
    if now != _last_time:
        _last_time = now
        time_changed = True
        changed_set.add('time')

    changed_set.add('time_ms')

    for port in ports.all_ports():
        if not port.is_enabled():
            continue

        port.invalidate_attrs()
        old_value = port.get_value()

        try:
            port.heart_beat()

        except Exception as e:
            logger.error('port heart beat exception: %s', e, exc_info=True)

        if time_changed:
            try:
                port.heart_beat_second()

            except Exception as e:
                logger.error('port heart beat second exception: %s', e, exc_info=True)

        try:
            new_value = await port.read_transformed_value()

        except Exception as e:
            logger.error('failed to read value from %s: %s', port, e, exc_info=True)

            continue

        if new_value is None:
            continue  # read value not available

        if new_value != old_value:
            old_value_str = json_utils.dumps(old_value) if old_value is not None else '(unavailable)'
            new_value_str = json_utils.dumps(new_value)

            logger.debug('detected %s value change: %s -> %s', port, old_value_str, new_value_str)

            port._value = new_value
            changed_set.add(port)

            # remember and reset port change reason
            change_reasons[port] = port.get_change_reason()
            port.reset_change_reason()

    if changed_set:
        await handle_value_changes(changed_set, change_reasons)

    sessions.respond_non_empty()
    sessions.cleanup()


async def update_loop():
    while _running:
        try:
            await update()

        except Exception as e:
            logger.error('update failed: %s', e, exc_info=True)

        await asyncio.sleep(settings.core.tick_interval / 1000.0)


async def handle_value_changes(changed_set, change_reasons):
    global _force_eval_expressions

    from . import expressions
    from . import ports

    if _force_eval_expressions:
        changed_set.add(None)  # special "always depends on" value
        _force_eval_expressions = False

    # trigger value-change events; save persisted ports
    for port in changed_set:
        if not isinstance(port, ports.BasePort):
            continue

        port.trigger_value_change()

        if await port.is_persisted():
            await port.save()

    # reevaluate the expressions depending on changed ports
    for port in ports.all_ports():
        if not port.is_enabled():
            continue

        expression = await port.get_expression()
        if expression:
            deps = expression.get_deps()
            deps.add(None)  # special "always depends on" value

            # if port expression depends on port itself and the change reason is the evaluation of its expression,
            # prevent evaluating its expression again to avoid evaluation loops

            change_reason = change_reasons.get(port, ports.CHANGE_REASON_NATIVE)
            if ((port in changed_set) and
                (('${}'.format(port.get_id())) in deps) and
                (change_reason == ports.CHANGE_REASON_EXPRESSION)):
                continue

            changed_set_ids = set('${}'.format(c.get_id()) for c in changed_set if isinstance(c, ports.BasePort))

            # evaluate a port's expression when:
            # * one of its deps changed
            # * its own value changed

            # join all deps together; deps may contain:
            # * ports
            # * port ids
            # * time dep strings
            # * None
            if not (deps & (changed_set_ids | changed_set)) and (port not in changed_set):
                continue

            try:
                value = expression.eval()

            except expressions.IncompleteExpression:
                continue

            except Exception as e:
                logger.error('failed to evaluate expression "%s" of %s: %s', expression, port, e)
                continue

            value = await port.adapt_value_type(value)
            if value is None:
                continue

            if value != port.get_value():
                logger.debug('expression "%s" of %s evaluated to %s', expression, port, json_utils.dumps(value))
                port.push_value(value, reason=ports.CHANGE_REASON_EXPRESSION)


def force_eval_expressions():
    global _force_eval_expressions

    _force_eval_expressions = True


def is_ready():
    return _ready


def set_ready():
    global _ready

    logger.debug('ready')
    _ready = True


async def init():
    global _update_loop_task

    force_eval_expressions()
    _update_loop_task = loop.create_task(update_loop())


async def cleanup():
    global _running

    _running = False

    await _update_loop_task




import pytz


def from_utc(moment, timezone):
    # noinspection PyTypeChecker
    if isinstance(timezone, str):
        timezone = pytz.timezone(timezone)
    
    if moment.tzinfo is None:
        moment = moment.replace(tzinfo=pytz.UTC)
    
    return moment.astimezone(timezone)


def to_utc(moment, timezone=pytz.UTC):
    if moment.tzinfo is None:
        # noinspection PyTypeChecker
        if isinstance(timezone, str):
            moment = pytz.timezone(timezone).localize(moment)

        else:
            moment = moment.replace(tzinfo=timezone)

    return moment.astimezone(pytz.UTC)




import logging.handlers


class FifoMemoryHandler(logging.handlers.MemoryHandler):
    def __init__(self, capacity):
        logging.handlers.MemoryHandler.__init__(self, capacity, flushLevel=logging.FATAL + 1)

    def flush(self):
        self.acquire()

        while len(self.buffer) > self.capacity:
            self.buffer.pop(0)

        self.release()




import abc
import logging
import time

from qtoggleserver.core import api as core_api


logger = logging.getLogger(__package__)


class Event(metaclass=abc.ABCMeta):
    REQUIRED_ACCESS = core_api.ACCESS_LEVEL_NONE
    TYPE = 'base-event'

    def __init__(self, timestamp=None):
        self._type = self.TYPE
        self._timestamp = timestamp or time.time()

    def __str__(self):
        return '{} event'.format(self._type)

    async def to_json(self):
        return {
            'type': self._type,
            'params': await self.get_params()
        }

    async def get_params(self):
        return {}

    def get_type(self):
        return self._type

    def get_timestamp(self):
        return self._timestamp

    def is_duplicate(self, event):
        return False




from qtoggleserver.core import api as core_api
from qtoggleserver.core import responses as core_responses


class SlaveError(Exception):
    pass


class InvalidDevice(SlaveError):
    pass


class NoListenSupport(SlaveError):
    def __init__(self, name):
        self.name = name

        super().__init__('device {} has no listen support'.format(name))


class DeviceAlreadyExists(SlaveError):
    def __init__(self, name):
        self.name = name

        super().__init__('device {} already exists'.format(name))


class DeviceRenamed(SlaveError):
    def __init__(self, slave):
        self.slave = slave

        super().__init__('{} renamed'.format(slave))


class DeviceOffline(SlaveError):
    def __init__(self, slave):
        self.slave = slave

        super().__init__('{} is offline'.format(slave))


class PortNotFound(SlaveError):
    def __init__(self, slave, _id):
        self.slave = slave
        self.id = _id

        super().__init__('could not find port {}.{}'.format(slave, _id))


def adapt_api_error(error):
    if isinstance(error, (core_responses.HostUnreachable,
                          core_responses.NetworkUnreachable,
                          core_responses.UnresolvableHostname)):

        return core_api.APIError(502, 'unreachable')

    elif isinstance(error, core_responses.ConnectionRefused):
        return core_api.APIError(502, 'connection refused')

    elif isinstance(error, core_responses.InvalidJson):
        return core_api.APIError(502, 'invalid device')

    elif isinstance(error, core_responses.Timeout):
        return core_api.APIError(504, 'device timeout')

    elif isinstance(error, core_responses.HTTPError):
        return core_api.APIError.from_http_error(error)

    elif isinstance(error, DeviceOffline):
        return core_api.APIError(503, 'device offline')

    elif isinstance(error, InvalidDevice):
        return core_api.APIError(502, 'invalid device')

    elif isinstance(error, core_responses.AuthError):
        return core_api.APIError(400, 'forbidden')  # yes, 400, since it's a slave authorization issue

    elif isinstance(error, core_api.APIError):
        return error

    else:  # Leave error unchanged since it's probably an internal exception
        return error




import abc
import asyncio
import copy
import logging

from .peripheral import Peripheral, PeripheralPort


READ_INTERVAL_ATTRDEF = {
    'display_name': 'Read Interval',
    'description': 'How often to read peripheral data (set to 0 to disable reading).',
    'type': 'number',
    'modifiable': True,
    'integer': True
}


logger = logging.getLogger(__name__)


class PolledPeripheral(Peripheral):
    DEFAULT_POLL_INTERVAL = 1800

    def __init__(self, address, name, **kwargs):
        self._polling = False
        self._poll_stopped = False
        self._poll_task = None
        self._poll_interval = self.DEFAULT_POLL_INTERVAL

        super().__init__(address, name, **kwargs)

    async def _poll_loop(self):
        self.debug('polling started')
        self._polling = True

        while self._polling:
            if self._poll_interval == 0:
                await asyncio.sleep(1)
                continue

            await self.poll()

            # Granular sleep so it can be interrupted
            orig_poll_interval = self._poll_interval
            for i in range(self._poll_interval):
                if not self._polling:
                    break

                if orig_poll_interval != self._poll_interval:  # Poll interval changed
                    break

                await asyncio.sleep(1)

        self._poll_task = None
        self.debug('polling stopped')

    def set_poll_interval(self, interval):
        self._poll_interval = interval

    def get_poll_interval(self):
        return self._poll_interval

    async def poll(self):
        raise NotImplementedError

    async def handle_enable(self):
        self._poll_task = asyncio.create_task(self._poll_loop())

    async def handle_disable(self):
        self._polling = False  # Will stop poll loop

    async def handle_cleanup(self):
        self._polling = False
        return self._poll_task


class PolledPort(PeripheralPort, metaclass=abc.ABCMeta):
    PERIPHERAL_CLASS = PolledPeripheral

    # Set these to None to disable read interval attribute
    READ_INTERVAL_MIN = 0
    READ_INTERVAL_MAX = 86400
    READ_INTERVAL_STEP = 1
    READ_INTERVAL_MULTIPLIER = 1
    READ_INTERVAL_UNIT = None

    def __init__(self, address, peripheral_name=None, **kwargs):
        super().__init__(address, peripheral_name, **kwargs)

        # Add read interval attrdef
        if self.READ_INTERVAL_MIN is not None:
            attrdef = copy.deepcopy(READ_INTERVAL_ATTRDEF)

            unit = self.READ_INTERVAL_UNIT
            if unit is None:
                if self.READ_INTERVAL_MULTIPLIER == 3600:
                    unit = 'hours'

                elif self.READ_INTERVAL_MULTIPLIER == 60:
                    unit = 'minutes'

                else:
                    unit = 'seconds'

            attrdef.update(unit=unit, step=self.READ_INTERVAL_STEP,
                           min=self.READ_INTERVAL_MIN, max=self.READ_INTERVAL_MAX)

            self.ADDITIONAL_ATTRDEFS = dict(self.ADDITIONAL_ATTRDEFS, read_interval=attrdef)

    async def attr_set_read_interval(self, interval):
        self.get_peripheral().set_poll_interval(int(interval) * self.READ_INTERVAL_MULTIPLIER)

    async def attr_get_read_interval(self):
        return self.get_peripheral().get_poll_interval() / self.READ_INTERVAL_MULTIPLIER




import asyncio
import logging
import re
import types

from tornado import httputil
from tornado.httpclient import AsyncHTTPClient, HTTPRequest

from qtoggleserver import persist
from qtoggleserver.conf import settings
from qtoggleserver.core import api as core_api
from qtoggleserver.core import responses as core_responses
from qtoggleserver.core.api import auth as core_api_auth
from qtoggleserver.utils import http as http_utils
from qtoggleserver.utils import json as json_utils


# a list of API calls that are prohibited via reverse mechanism
BLACKLIST_CALLS = [
    ('GET', re.compile(r'^/listen'))
]

logger = logging.getLogger(__name__)
_reverse = None


class InvalidParamError(Exception):
    def __init__(self, param):
        self._param = param
        super().__init__('invalid field: {}'.format(param))


class InvalidConsumerRequestError(Exception):
    pass


class UnauthorizedConsumerRequestError(Exception):
    pass


class Reverse:
    # noinspection PyUnusedLocal
    def __init__(self, scheme=None, host=None, port=None, path=None, device_id=None, password=None, timeout=None,
                 **kwargs):

        # the enabled value comes with kwargs but is ignored,
        # as the reverse object will be explicitly enabled afterwards

        self._scheme = scheme
        self._host = host
        self._port = port
        self._path = path
        self._device_id = device_id
        self._password = password
        self._timeout = timeout

        self._enabled = False
        self._url = None

    def __str__(self):
        s = 'reverse'

        if not self._enabled:
            s += ' [disabled]'

        if self._scheme:
            s += ' ' + self.get_url()

        return s

    def get_url(self):
        if not hasattr(self, '_url'):
            if self._scheme == 'http' and self._port == 80 or self._scheme == 'https' and self._port == 443:
                self._url = '{}://{}{}'.format(self._scheme, self._host, self._path)

            else:
                self._url = '{}://{}:{}{}'.format(self._scheme, self._host, self._port, self._path)

        return self._url

    def is_enabled(self):
        return self._enabled

    def enable(self):
        if self._enabled:
            return

        logger.debug('starting wait loop')

        self._enabled = True
        asyncio.create_task(self._session_loop())

    def disable(self):
        if not self._enabled:
            return

        logger.debug('wait stopped')

        self._enabled = False

    def to_json(self):
        d = {
            'enabled': self._enabled,
            'scheme': self._scheme,
            'host': self._host,
            'port': self._port,
            'path': self._path,
            'device_id': self._device_id,
            'timeout': self._timeout
        }

        return d

    async def _session_loop(self):
        api_response_dict = None
        sleep_interval = 0

        while True:
            if not self._enabled:
                break

            if sleep_interval:
                await asyncio.sleep(sleep_interval)
                sleep_interval = 0

            try:
                api_request_dict = await self._wait(api_response_dict)  # TODO properly implement me

            except UnauthorizedConsumerRequestError:
                api_response_dict = {
                    'status': 401,
                    'body': json_utils.dumps({'error': 'authentication required'})
                }

                continue

            except Exception as e:
                logger.error('wait failed: %s, retrying in %s seconds', e, settings.reverse.retry_interval,
                             exc_info=True)
                sleep_interval = settings.reverse.retry_interval
                continue

            # the reverse mechanism has been disabled while waiting
            if not self._enabled:
                break

            try:
                api_response_dict = await self._process_api_request(api_request_dict)

            except Exception as e:
                logger.error('reverse API call failed: %s', e, exc_info=True)
                sleep_interval = settings.reverse.retry_interval
                api_response_dict = None
                continue

    async def _wait(self, api_request_dict, api_response_dict):
        url = self.get_url()

        headers = {
            'Content-Type': http_utils.JSON_CONTENT_TYPE,
            'Authorization': core_api_auth.make_auth_header(core_api_auth.ORIGIN_DEVICE,
                                                            username=self._device_id,
                                                            password_hash=self._password)
        }

        body_str = None
        if api_response_dict:  # answer request
            body_str = api_response_dict['body']
            headers['Status'] = '{} {}'.format(api_response_dict['status'],
                                               httputil.responses[api_response_dict['status']])
            headers['API-Call-Id'] = api_request_dict['api_call_id']

        http_client = AsyncHTTPClient()
        request = HTTPRequest(url, 'POST', headers=headers, body=body_str,
                              connect_timeout=self._timeout, request_timeout=self._timeout)

        if api_response_dict:
            logger.debug('sending answer request to %s %s (API call id %s) to %s',
                         api_request_dict['method'], api_request_dict['path'], api_request_dict['api_call_id'], url)

        else:
            logger.debug('sending initial request to %s', url)

        try:
            # this response is in fact an API request
            consumer_response = await http_client.fetch(request, raise_error=False)

        except Exception as e:
            # We need to catch exceptions here even though raise_error is False,
            # because it only affects HTTP errors
            consumer_response = types.SimpleNamespace(error=e, code=599)

        api_request_dict = self._parse_consumer_response(consumer_response)

        return api_request_dict

    @staticmethod
    def _parse_consumer_response(response):
        body = core_responses.parse(response)  # will raise for non-2xx

        auth = response.headers.get('Authorization')
        if not auth:
            raise UnauthorizedConsumerRequestError('missing authorization header')

        try:
            usr = core_api_auth.parse_auth_header(auth, core_api_auth.ORIGIN_CONSUMER,
                                                  core_api_auth.consumer_password_hash_func)

        except core_api_auth.AuthError as e:
            raise UnauthorizedConsumerRequestError(str(e)) from e

        access_level = core_api.ACCESS_LEVEL_MAPPING[usr]

        try:
            method = response.headers['Method']

        except KeyError:
            raise InvalidConsumerRequestError('missing Method header') from None

        try:
            path = response.headers['Path']

        except KeyError:
            raise InvalidConsumerRequestError('missing Path header') from None

        try:
            api_call_id = response.headers['API-Call-Id']

        except KeyError:
            raise InvalidConsumerRequestError('missing API-Call-Id header') from None

        return {
            'body': body,
            'method': method,
            'path': path,
            'api_call_id': api_call_id,
            'access_level': access_level,
            'username': usr
        }

    async def _process_api_request(self, request_dict):
        from qtoggleserver.web import server as web_server

        if self._is_black_listed(request_dict):
            return {
                'status': 404,
                'body': json_utils.dumps({'error': 'no such function'})
            }

        request = httputil.HTTPServerRequest(method=request_dict['method'],
                                             uri=request_dict['path'],
                                             body=request_dict['body'])

        dispatcher = web_server.get_application().find_handler(request)
        dispatcher.execute()

        status = dispatcher.handler.get_status()
        body = dispatcher.handler.get_response_body()

        return {
            'status': status,
            'body': body
        }

    @staticmethod
    def _is_black_listed(request_dict):
        for method, path_re in BLACKLIST_CALLS:
            if request_dict['method'] != method:
                continue

            if path_re.match(request_dict['path']):
                return True

        return False


def get():
    return _reverse


def setup(enabled, scheme=None, host=None, port=None, path=None, device_id=None, password=None, timeout=None, **kwargs):
    global _reverse

    if _reverse and _reverse.is_enabled():
        _reverse.disable()

    _reverse = Reverse(scheme, host, port, path, device_id, password, timeout)
    if enabled:
        _reverse.enable()


def load():
    data = persist.get_value('reverse')
    if data is None:
        setup(enabled=False)
        logger.debug('loaded %s', _reverse)

        return

    setup(**data)
    logger.debug('loaded %s', _reverse)


def save():
    if _reverse is None:
        return

    logger.debug('saving data')
    persist.set_value('reverse', _reverse.to_json())


def reset():
    logger.debug('clearing persisted data')
    persist.remove('reverse')
